#!/usr/bin/env python3

import aiocells


def hello_world():
    print("Hello, world!")


def main():
    graph = aiocells.DependencyGraph()

    # The node can be any callable, in this case a function.
    graph.add_node(hello_world)
    aiocells.compute_sequential(graph)



#!/usr/bin/env python3

import time

import aiocells


def main():
    graph = aiocells.DependencyGraph()

    # 'add_node' always returns the node that has just been added, in this
    # case the lambda functions. We will use this below to define precedence
    # relationships
    print_sleeping = graph.add_node(lambda: print("Sleeping..."))
    sleep = graph.add_node(lambda: time.sleep(2))
    print_woke_up = graph.add_node(lambda: print("Woke up!"))

    print("Define the precedence relationships...")
    graph.add_precedence(print_sleeping, sleep)
    graph.add_precedence(sleep, print_woke_up)

    # Now, after we've defined the precedence relationships, we use the
    # simplest computer to compute the graph. The nodes will be called in
    # an order that is consistent with the precedence relationships.
    # Specifically, the nodes are executed in topological order.
    aiocells.compute_sequential(graph)



import asyncio
import inspect
import logging

import aiocells.aio as aio

REPEATER = "cells.flow.repeater"

logger = logging.getLogger(__name__)


def repeat(function):
    if not inspect.iscoroutinefunction(function):
        raise ValueError("Event source must be a coroutine function")
    setattr(function, REPEATER, True)
    return function


def is_repeater(function):
    return getattr(function, REPEATER, False)


async def compute_flow(graph):
    running_tasks = []
    try:
        assert graph is not None
        input_nodes = graph.input_nodes
        callables, running_tasks = aio.prepare_ready_set(input_nodes)
        if len(callables) > 0:
            raise Exception("Input nodes must be coroutines")
        # Wait for at least one input node to complete
        while len(running_tasks) > 0:
            logger.debug("Waiting for input tasks")
            completed_tasks, running_tasks = await asyncio.wait(
                running_tasks,
                return_when=asyncio.FIRST_COMPLETED
            )
            logger.debug("Input received")
            aio.raise_task_exceptions(completed_tasks)
            completed_repeater_functions = [
                task.aio_coroutine_function
                for task in completed_tasks
                if is_repeater(task.aio_coroutine_function)
            ]
            callables, new_tasks = aio.prepare_ready_set(
                completed_repeater_functions
            )
            assert len(callables) == 0
            running_tasks |= new_tasks
            logger.debug("Computing dependent nodes")
            for node in graph.topological_ordering:
                if node in input_nodes:
                    continue
                logger.debug("Computing dependent node: %s", node)
                if inspect.iscoroutinefunction(node):
                    await node()
                else:
                    assert callable(node)
                    node()
    except asyncio.CancelledError as e:
        await aio.cancel_tasks(running_tasks)
        raise
    except Exception as e:
        await aio.cancel_tasks(running_tasks)
        logger.exception("compute_flow error")
        raise



#!/usr/bin/env python3

import asyncio
import datetime
import functools
import logging

import aiocells


logger = logging.getLogger()


# Demonstrates a node where the function is actually a coroutine method.
# In this case, functools.partial must be used to bind the function and
# the object.

class TimerObject:

    def __init__(self, result_variable):
        self.result_variable = result_variable

    async def compute(self):
        logger.debug("sleeping")
        await asyncio.sleep(1)
        logger.debug("woke up")
        self.result_variable.value = datetime.datetime.now()


def main():

    graph = aiocells.DependencyGraph()

    time = aiocells.Variable()
    timer = TimerObject(time)

    # Here, we bind the coroutine method with the object
    compute_timer = functools.partial(TimerObject.compute, timer)
    printer = aiocells.print_value(time, "variable changed to {value}")

    graph.add_precedence(compute_timer, time)
    graph.add_precedence(time, printer)
    logger.debug("graph: %s", graph)

    logger.info("First computation...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    logger.debug("graph: %s", graph)

    logger.info("Second computation...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    logger.debug("graph: %s", graph)




from .basic import DependencyGraph, Variable, compute_sequential, Stopwatch, \
        print_value, assign

from .aio import async_compute_sequential, async_compute_concurrent, \
        async_compute_concurrent_simple, timer

from .mod import ModClock, ModVariable, ModPrinter

from .flow import compute_flow, repeat



#!/usr/bin/env python3

import asyncio
from functools import partial

import aiocells


def create_graph(stopwatch):

    graph = aiocells.DependencyGraph()

    start_stopwatch = stopwatch.start
    stop_stopwatch = stopwatch.stop

    for t in range(100000):
        sleep_1 = partial(asyncio.sleep, 1)
        graph.add_precedence(start_stopwatch, sleep_1)
        graph.add_precedence(sleep_1, stop_stopwatch)

    return graph


def main():

    stopwatch = aiocells.Stopwatch()
    graph = create_graph(stopwatch)

    # How long does it take to run 100000 async 1 second sleeps with
    # async_compute_concurrent_simple?
    print("Running 100000 async 1 second sleeps with"
          " async_compute_concurrent_simple...")
    asyncio.run(aiocells.async_compute_concurrent_simple(graph))
    print("Computation with `async_compute_concurrent_simple` took"
          f" {stopwatch.elapsed_time()}")

    # How long does it take to run 100000 async 1 second sleeps with
    # async_compute_concurrent?
    print("Running 100000 async 1 second sleeps with"
          " async_compute_concurrent...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    print("Computation with `async_compute_concurrent` took"
          f" {stopwatch.elapsed_time()}")



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


def main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    time = aiocells.ModVariable(clock)
    printer = aiocells.ModPrinter(clock, time, "time changed to {value}")
    graph.add_precedence(time, printer)

    timer_0 = functools.partial(aiocells.timer, 0, time)
    graph.add_precedence(timer_0, time)

    # Flow graphs are generally interested in input values that keep changing
    # over time. Here, we simulate that by setting up a repeating timer. We do
    # that, with the 'aiocells.repeat' function. This function marks the passed
    # function as a 'repeater function', which instructs
    # 'aiocells.compute_flow' to repeat the function every time it returns.
    # This, the timer will be repeated ad infinitum.
    #
    # This example will continue until it is interrupted with Ctrl-C.
    #
    # Note that marking a function to be repeater function currently only
    # affects the behaviour of 'compute_flow' and not any of the other
    # `compute` functions.

    print()
    print("Ctrl-C to exit the demo")
    print()
    repeat_timer = aiocells.repeat(functools.partial(aiocells.timer, 1, time))
    graph.add_precedence(repeat_timer, time)

    asyncio.run(aiocells.compute_flow(graph))



#!/usr/bin/env python3

import asyncio
from functools import partial

import aiocells

# Demonstrates adding nodes to a graph which themselves are graph
# computations. Importantly, the subgraphs execute concurrently with
# one another.


async def run_subgraph(name):
    graph = aiocells.DependencyGraph()
    sleep_2 = partial(asyncio.sleep, 2)
    graph.add_node(sleep_2)
    print(f"Running subgraph {name}")
    await aiocells.async_compute_concurrent(graph)


def create_graph(stopwatch):

    graph = aiocells.DependencyGraph()

    start_stopwatch = graph.add_node(stopwatch.start)
    stop_stopwatch = graph.add_node(stopwatch.stop)

    for i in range(10):
        subgraph = graph.add_node(partial(run_subgraph, f"{i}"))
        graph.add_precedence(start_stopwatch, subgraph)
        graph.add_precedence(subgraph, stop_stopwatch)

    return graph


def main():

    stopwatch = aiocells.Stopwatch()
    graph = create_graph(stopwatch)

    asyncio.run(aiocells.async_compute_concurrent(graph))
    print("Computation with async_compute_concurrent took"
          f" {stopwatch.elapsed_time()}")



#!/usr/bin/env python3

import asyncio
import time

import aiocells


def create_graph(stopwatch):
    graph = aiocells.DependencyGraph()

    start_stopwatch = stopwatch.start

    # Note that these two sleeps are just lambdas. They are vanilla
    # callables and not coroutine functions. Thus, they cannot be run
    # concurrently. Thus, the total run time will be about 3 seconds even
    # if we use the concurrent computer.
    sleep_1 = graph.add_node(lambda: time.sleep(1))
    sleep_2 = graph.add_node(lambda: time.sleep(2))

    stop_stopwatch = stopwatch.stop

    graph.add_precedence(start_stopwatch, sleep_1)
    graph.add_precedence(sleep_1, stop_stopwatch)

    graph.add_precedence(start_stopwatch, sleep_2)
    graph.add_precedence(sleep_2, stop_stopwatch)

    return graph


def main():

    stopwatch = aiocells.Stopwatch()
    graph = create_graph(stopwatch)

    print("Should take about 3 seconds...")
    asyncio.run(aiocells.async_compute_concurrent_simple(graph))
    print("Computation with `async_compute_concurrent_simple` took "
          f"{stopwatch.elapsed_time()}")

    print("Should take about 3 seconds...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    print("Computation with `async_compute_concurrent` took "
          f"{stopwatch.elapsed_time()}")



#!/usr/bin/env python3

import asyncio

import aiocells


def create_graph(stopwatch):

    graph = aiocells.DependencyGraph()

    start_stopwatch = stopwatch.start
    stop_stopwatch = stopwatch.stop

    for t in range(100000):
        def null():
            pass
        graph.add_precedence(start_stopwatch, null)
        graph.add_precedence(null, stop_stopwatch)

    return graph


def main():

    stopwatch = aiocells.Stopwatch()
    graph = create_graph(stopwatch)

    # How long does it take to compute 100000 null callables with
    # async_compute_concurrent_simple?
    asyncio.run(aiocells.async_compute_concurrent_simple(graph))
    print("Computation with async_compute_concurrent_simple took"
          f"{stopwatch.elapsed_time()}")

    # How long does it take to compute 100000 null callables with
    # async_compute_concurrent?
    asyncio.run(aiocells.async_compute_concurrent(graph))
    print("Computation with async_compute_concurrent took"
          f" {stopwatch.elapsed_time()}")



#!/usr/bin/env python3

import asyncio
from functools import partial

import aiocells


def create_graph(stopwatch):

    graph = aiocells.DependencyGraph()

    # The method to start the stopwatch
    start_stopwatch = stopwatch.start

    # Two sleeps. Note that they are asyncio.sleep
    sleep_1 = partial(asyncio.sleep, 1)
    sleep_2 = partial(asyncio.sleep, 2)

    # The method to stop the stopwatch
    stop_stopwatch = stopwatch.stop

    # Start the stopwatch before the first sleep
    graph.add_precedence(start_stopwatch, sleep_1)
    # Stop the stopwatch after the first sleep
    graph.add_precedence(sleep_1, stop_stopwatch)

    # Start the stopwatch before the second sleep
    graph.add_precedence(start_stopwatch, sleep_2)
    # Stop the stopwatch after the second sleep
    graph.add_precedence(sleep_2, stop_stopwatch)

    # Note that there is no precedence relationship between the two
    # sleeps.
    return graph


def main():

    stopwatch = aiocells.Stopwatch()
    graph = create_graph(stopwatch)
    # Even though the graph is a diamond (the sleeps do no depend on each
    # other and _could_ be executed concurrenty, `async_compute_sequential`
    # does not support concurrent execution. Thus, the execution time is
    # about 3 seconds, the sum of the two sleeps.
    print("Two async sleeps computed sequentially.")
    print("Total time should take about 3 seconds...")
    asyncio.run(aiocells.async_compute_sequential(graph))
    print("Computation with `async_compute_sequential` took"
          f" {stopwatch.elapsed_time()}")
